# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of shogen.
#
# shogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# shogen is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with shogen.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
import re
from pyasp.asp import *
import xml.etree.ElementTree as etree  
from xml.etree.ElementTree import XML, fromstring, tostring


    
def get_model(sbml):
    model_element = None
    for e in sbml:
        if e.tag[0] == "{":
	  uri, tag = e.tag[1:].split("}")
	else: tag = e.tag
        if tag == "model":
	  model_element = e
          break
    return model_element    
    
def get_listOfSpecies(model):
    listOfSpecies = None
    for e in model:
        if e.tag[0] == "{":
	  uri, tag = e.tag[1:].split("}")
	else: tag = e.tag
        if tag == "listOfSpecies":
	  listOfSpecies = e
          break
    return listOfSpecies 
    
def get_listOfReactions(model):
    listOfReactions = None
    for e in model:
        if e.tag[0] == "{":
	  uri, tag = e.tag[1:].split("}")
	else: tag = e.tag
        if tag == "listOfReactions":
	  listOfReactions = e
          break
    return listOfReactions 
    
def get_listOfReactants(reaction):
    listOfReactants = None
    for e in reaction:
        if e.tag[0] == "{":
	  uri, tag = e.tag[1:].split("}")
	else: tag = e.tag
        if tag == "listOfReactants":
	  listOfReactants = e
          break
    return listOfReactants
    
def get_listOfProducts(reaction):
    listOfProducts = None
    for e in reaction:
        if e.tag[0] == "{":
	  uri, tag = e.tag[1:].split("}")
	else: tag = e.tag
        if tag == "listOfProducts":
	  listOfProducts = e
          break
    return listOfProducts
 
def get_score(reaction):
    score = 0
    for e in reaction:
      if e.tag[0] == "{":
	uri, tag = e.tag[1:].split("}")
      else: tag = e.tag
      if tag == "annotation":
	 for s in e:
          if s.tag[0] == "{":
	    suri, stag = s.tag[1:].split("}")
	  else: stag = s.tag 
	  if stag == "score":
	    score = s.attrib.get("val")
    return score
    
    
def readSBMLnetwork(filename) :
  
   lpfacts = TermSet()
   
   tree = etree.parse(filename)
   sbml = tree.getroot()
   model = get_model(sbml)
   
   listOfSpecies = get_listOfSpecies(model)
   for e in listOfSpecies:
     if e.tag[0] == "{":
       uri, tag = e.tag[1:].split("}")
     else: tag = e.tag
     #if tag == "species":
       #lpfacts.add(Term('species', ["\""+e.attrib.get("id")+"\""]))
  
   listOfReactions = get_listOfReactions(model)
   for e in listOfReactions:
     if e.tag[0] == "{":
       uri, tag = e.tag[1:].split("}")
     else: tag = e.tag
     if tag == "reaction":
       #reactionId = e.attrib.get("id")
       reactionId = e.attrib.get("name")
       lpfacts.add(Term('reaction', ["\""+reactionId+"\""]))
       if(e.attrib.get("reversible")=="true"):  lpfacts.add(Term('reversible', ["\""+reactionId+"\""]))
       
       listOfReactants = get_listOfReactants(e)
       if listOfReactants== None : print "\n Warning:",reactionId, "listOfReactants=None"
       else: 
	  for r in listOfReactants:
	    lpfacts.add(Term('reactant', ["\""+r.attrib.get("species")+"\"", "\""+reactionId+"\""]))
         
       listOfProducts = get_listOfProducts(e)
       if listOfProducts== None : print "\n Warning:",reactionId, "listOfProducts=None"
       else: 
	  for p in listOfProducts:
	    lpfacts.add(Term('product', ["\""+p.attrib.get("species")+"\"", "\""+reactionId+"\""]))

   return lpfacts
   





                



# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of shogen.
#
# shogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# shogen is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with shogen.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
import os
from pyasp.asp import *



def read_genome(genome_string) :
  instance = TermSet()

  genome_f = open(genome_string, "r")
  genome = genome_f.read().replace('\r\n','\n').split()
  dictg = {} # Dictionnary for the genes
  revdictg = [0]
  g = 0 # Counter to define the matchings between real names and ASP names
  for line in genome :
    g+=1
    dictg[line] = g   
    revdictg.append(line)
    instance.add(Term('gene', [g])) 
  
  genome_f.close()
  return instance, dictg

def read_catalysis(catalysation_string, gene_dict) :
    
  instance = TermSet()
  catalysis_f = open(catalysation_string, "r")

  i=0
  catalyzis = catalysis_f.readlines()
  for line in catalyzis :
    cat = line.replace('\r\n','\n').split()
  
    if gene_dict.has_key(cat[1]) :
      instance.add(Term('cat', [gene_dict[cat[1]],"\""+cat[0]+"\""]))

  catalysis_f.close()
  return instance

  

def createInstance(genome_string,metabolism_string,catalysation_string) :

  instance = TermSet()

  igenome = open(genome_string, "r")
  genome = igenome.read().replace('\r\n','\n').split()
  dictg = {} # Dictionnary for the genes
  revdictg = [0]
  g = 0 # Counter to define the matchings between real names and ASP names
  for line in genome :
    g+=1
    dictg[line] = g   
    revdictg.append(line)
    instance.add(Term('gene', [g]))
    
  
  imetabolism = open(metabolism_string, "r")
  metabolism = imetabolism.read().replace('\r\n','\n').splitlines()
  dictr = {}# Dictionnary for the reactions
  revdictr = [0]
  r = 0 # Counter to define the matchings between real names and ASP names
  
  for line in metabolism :
    if line !="":
      link = line.split()
      if not dictr.has_key(link[0]) :
	r+=1
	dictr[link[0]] = r
	revdictr.append(link[0])
      if not dictr.has_key(link[1]) :
	r+=1
	dictr[link[1]] = r
	revdictr.append(link[1])
      instance.add(Term('redge', [dictr[link[0]],dictr[link[1]]]))

  #print("\t\t</nodes>\n")

  
  icatalyzis = open(catalysation_string, "r")

  catalyzistab=[]
  i=0
  catalyzis = icatalyzis.readlines()
  for line in catalyzis :
    cat = line.replace('\r\n','\n').split()
  
    if dictr.has_key(cat[0]) and dictg.has_key(cat[1]) :
      instance.add(Term('cat', [dictg[cat[1]],dictr[cat[0]]]))
      while i < dictr[cat[0]]+1 :
	catalyzistab.append([])
	i+=1
      catalyzistab[dictr[cat[0]]].append(dictg[cat[1]])  
    #else :
      #print "irrelevant gene or reaction",cat[1],cat[0]

  #print len(catalyzistab)      
  #exit(0)
      
  couplestab = []
  for i in range(len(catalyzistab)) :
    for j in range(len(catalyzistab))[(i+1):] :
      condition = True
      for g1 in catalyzistab[i] :
	for g2 in catalyzistab[j] :
	  if condition :
	    glen = len(dictg)
	    if min(abs(g1-g2),glen-abs(g1-g2)) <= 10 :
	      couplestab.append([i,j])
	      condition = False    
	      
  return instance, revdictg, revdictr, dictr, couplestab


def readcouples(couple_string, dictr, revdictr) :

  icouples = open(couple_string, "r")
  bla =  icouples.read().replace('\r\n','\n').splitlines()
  r = len(revdictr) # Counter to define the matchings between real names and ASP names
  couples = TermSet()
  for line in bla :
    link = line.split()
    if not dictr.has_key(link[0]) :# Dictionnary for the reactions
      r+=1
      dictr[link[0]] = r 
      revdictr.append(link[0])
    if not dictr.has_key(link[1]) :
      r+=1
      dictr[link[1]] = r
      revdictr.append(link[1])
    couples.add(Term('pair', [str(dictr[link[0]]),str(dictr[link[1]])]))
  return couples, revdictr

def readcouples_new(couple_string) :
  couples = TermSet()
  
  couples_f = open(couple_string, "r")
  bla =  couples_f.read().replace('\r\n','\n').splitlines()

  for line in bla :
    link = line.split()
    couples.add(Term('pair', ["\""+link[0]+"\"","\""+link[1]+"\""]))
    
  couples_f.close()
  return couples


def print_reactions(answer,dictr,dictg) :
  for a in answer:
    if a.pred() == "fedge" :
      print "\n   ",dictr[int(a.arg(0).arg(1))]+"("+dictg[int(a.arg(0).arg(0))]+")",
      print"->",
      print dictr[int(a.arg(1).arg(1))]+"("+dictg[int(a.arg(1).arg(0))]+")",
      print ": ",a.arg(2),
  print

  
def print_genes(answer, dictg) :
  for a in answer:
    if a.pred() == "cgene" :
      print dictg[int(a.arg(0))],
    if a.pred() == "ugene" :
      print dictg[int(a.arg(0))],
  print
  
   
  
def clean_up() :
  if os.path.isfile("parser.out"): os.remove("parser.out")
  if os.path.isfile("asp_py_lextab.py"): os.remove("asp_py_lextab.py")
  if os.path.isfile("asp_py_lextab.pyc"): os.remove("asp_py_lextab.pyc")
  if os.path.isfile("asp_py_parsetab.py"): os.remove("asp_py_parsetab.py")
  if os.path.isfile("asp_py_parsetab.pyc"): os.remove("asp_py_parsetab.pyc") 


# Copyright (c) 2012, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of shogen.
#
# shogen is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# shogen is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with shogen.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
import os
import tempfile
from pyasp.asp import *

root = __file__.rsplit('/', 1)[0]

prepro_prg =      root + '/encodings/preprocess.lp'
length_prg =    root + '/encodings/sgs.lp'
convert_prg =   root + '/encodings/convert.lp'
convert_sds_prg =   root + '/encodings/convert_sds.lp'
filter_prg =      root + '/encodings/filter_couples.lp'


class ASPrinter:
  def write(self,count, termset):
    print count,":",termset
  
class EdgePrinter:
  def __init__(self,dictg,revdictr):
    self.dictg = dictg
    self.revdictr = revdictr
  def write(self,count, termset):
    print str(count)+":",
    for t in termset: 
      if t.pred() == "fedge" :
        print "("+self.dictg[int(t.arg(0).arg(0))]+","+self.revdictr[int(t.arg(0).arg(1))]+")-"+str(t.arg(2))+"->("+self.dictg[int(t.arg(1).arg(0))]+","+self.revdictr[int(t.arg(1).arg(1))]+")",
      if t.pred() == "zwoop" :
	print "zwoop("+self.dictg[int(t.arg(0).arg(0))]+","+self.revdictr[int(t.arg(0).arg(1))]+")->("+self.dictg[int(t.arg(1).arg(0))]+","+self.revdictr[int(t.arg(1).arg(1))]+")",
    print " "

    
class GenePrinter:
  def __init__(self,dictg,revdictr):
    self.dictg = dictg
    self.revdictr = revdictr
  def write(self,count, termset):
    print str(count)+":",
    for t in termset: 
      if t.pred() == "ugene" :
	print self.dictg[int(t.arg(0))],
      else:
	print t  
    print " "

    
def filter_couples(couple_facts, instance):
    prg = [filter_prg, instance, couple_facts.to_file()]
    solver = GringoClasp()
    models = solver.run(prg,collapseTerms=True, collapseAtoms=False)
    os.unlink(prg[2])
    return models[0]
  
   
def get_ksip_instance(instance, pmax):

    pmaxfact = String2TermSet('pmax('+str(pmax)+')')
  
    inst=pmaxfact.to_file() 
    prg = [convert_prg , instance, inst ]

    solver = GringoClasp()
    solution = solver.run(prg,collapseTerms=False, collapseAtoms=False)
    os.unlink(inst)
    return solution[0] 

    
def get_sgs_instance(instance, pmax):

    pmaxfact = String2TermSet('pmax('+str(pmax)+')')
  
    inst=pmaxfact.to_file() 
    prg = [convert_sds_prg , instance, inst ]

    solver = GringoClasp()
    solution = solver.run(prg,collapseTerms=False, collapseAtoms=False)
    os.unlink(inst)
    return solution[0] 
  
    
def preprocess(instance, start, goal, pmax):
    startfact = String2TermSet('start('+str(start)+')')
    goalfact = String2TermSet('goal('+str(goal)+')')
    pmaxfact = String2TermSet('pmax('+str(pmax)+')')
    details = startfact.union(goalfact).union(pmaxfact)
    details_f = details.to_file() 
    prg = [prepro_prg, instance, details_f ]

    solver = GringoClasp()
    solution = solver.run(prg,1)
    
    os.unlink(details_f)
    return solution[0] 
    
    
def get_k_sgs(instance, start, end, pmax, k, dictg, revdictr):
    startfact = String2TermSet('start('+str(start)+')')
    endfact = String2TermSet('end('+str(end)+')')
    pmaxfact = String2TermSet('pmax('+str(pmax)+')')
    
    details = startfact.union(endfact).union(pmaxfact)

    
    details_f=details.to_file() 
    count=0
    min=1
    geneprinter = GenePrinter(dictg, revdictr)
    
    #while len(solutions) < k :
    while count < k : 
      prg = [length_prg , instance, details_f ]
      goptions=' --const pmin='+str(min)
      coptions='--opt-heu --opt-strategy=1 --heu=vsids '
      #coptions='--opt-heu --heu=vsids'
      solver = GringoClasp(gringo_options=goptions,clasp_options=coptions)
      #solver = GringoUnClasp(gringo_options=goptions,clasp_options=coptions)
      #print "search1 ...",
      optima = solver.run(prg,collapseTerms=True,collapseAtoms=False)
   
      
      if len(optima) :
	count  += 1
	min= optima[0].score[0]
	
	print "length:",min
	geneprinter.write(count,optima[0])
	#min=optima[0][0]
	
	prg = [length_prg , instance, details_f, exclude_sol([optima[0]]) ]
	goptions='--const pmin='+str(min)
	coptions='--opt-heu --opt-strategy=1 --opt-mode=optN --opt-bound='+str(min)
	solver = GringoClasp(gringo_options=goptions,clasp_options=coptions)
	#print "search2 ...",
	sols = solver.run(prg,collapseTerms=True,collapseAtoms=False)  
	for i in sols :
	  count+=1
	  geneprinter.write(count,i)
	os.unlink(prg[3])
	min+=1
      else :
        os.unlink(details_f)
	return  
	
    os.unlink(details_f)
    return 

  